'use strict'

const { SESSION_WARNING } = require('../messages')
const BaseGate = require('../base_gate')
const RealmError = require('../realm_error').RealmError
const { mqttParse, restoreUri } = require('../topic_pattern')
const Context = require('../context')

/*
// mosquitto_pub -d -u realm1 -P passwd -t com.myapp.topic1 -m "the message"

parse Packet {
  cmd: 'connect',
  retain: false,
  qos: 0,
  dup: false,
  length: 49,
  topic: null,
  payload: null,
  protocolId: 'MQIsdp',
  protocolVersion: 3,
  clean: true,
  keepalive: 60,
  clientId: 'mosqpub|2386-ats-ora5',
  username: 'user',
  password: <Buffer 70 61 73 73 77 64> }
generate { returnCode: 0, sessionPresent: false, cmd: 'connack' }
parse Packet {
  cmd: 'publish',
  retain: false,
  qos: 0,
  dup: false,
  length: 17,
  topic: 'test',
  payload: <Buffer 74 68 65 20 6d 65 73 73 61 67 65> }
parse Packet {
  cmd: 'disconnect',
  retain: false,
  qos: 0,
  dup: false,
  length: 0,
  topic: null,
  payload: null }

===========================================================

$ mosquitto_sub -c -i this-agent-session --will-topic wt --will-payload wt-disconnected -t test

MQTT ARRIVED Packet {
  cmd: 'connect',
  retain: false,
  qos: 0,
  dup: false,
  length: 51,
  topic: null,
  payload: null,
  protocolId: 'MQTT',
  protocolVersion: 4,
  will: 
   { retain: false,
     qos: 0,
     topic: 'wt',
     payload: <Buffer 77 74 2d 64 69 73 63 6f 6e 6e 65 63 74 65 64> },
  clean: false,
  keepalive: 60,
  clientId: 'this-agent-session' }
 
 [1993019693010936] command not found Packet {
  cmd: 'puback',
  retain: false,
  qos: 0,
  dup: false,
  length: 2,
  topic: null,
  payload: null,
  messageId: 1 }

  */

let handlers = {}
let cmdAck = {}

class MqttContext extends Context {
  sendEvent (cmd) {
    let payload = ''
    if (cmd.data.payload !== undefined) {
      payload = cmd.data.payload
    } else
    if (cmd.data.args !== undefined) {
      payload = Buffer.from(JSON.stringify(cmd.data.args))
    } else
    if (cmd.data.kwargs !== undefined) {
      payload = Buffer.from(JSON.stringify(cmd.data.kwargs))
    } else
    if (cmd.data.kv !== undefined) {
      payload = Buffer.from(JSON.stringify(cmd.data.kv))
    }
    this.mqttSend({
      topic: restoreUri(cmd.uri),
      payload: payload,
      qos: (cmd.opt.history ? 1 : 0),
      messageId: this.getSession().genSessionMsgId(),
      cmd: 'publish'
    })
  }

  acknowledged (cmd) {
    cmdAck[cmd.mtype].call(this, cmd)
  }

  mqttSend (msg, callback) {
    this.sender.send(msg, callback)
  }

  mqttClose (code, reason) {
    this.sender.close(code, reason)
  }

  error(id, code, msg) {
    if (undefined === id) {
      throw new RealmError(this.getId(), code, msg)
    }
    else {
      // subscribe mode
      let pkg = this.getId()
      pkg.granted[id] = 0x80
      this.mqttSubscribeDone()
    }
  }

  mqttSubscribeDone () {
    // granted is accepted QoS for the subscription, possible [0,1,2, + 128 or 0x80 for failure]
    let pkg = this.getId()
    pkg.count--
    if (pkg.count === 0) {
      this.mqttSend({ cmd: 'suback', messageId: pkg.id, granted: pkg.granted })
    }
  }
}

class MqttGate extends BaseGate {
  createContext (session, sender) {
    return new MqttContext(this._router, session, sender)
  }

  checkRealm (session, requestId) {
    if (!session.realm) {
      throw new RealmError(requestId, 'wamp.error.not_authorized')
    }
  }

  hello (ctx, session, message) {
    session.realmName = message.username

    // session.setSessionName(message.clientId, message.clean)

    if (this.isAuthRequired(session)) {
      session.secureDetails = message
      if (details.hasOwnProperty('authmethods') && details.authmethods.indexOf('ticket') >= 0) {
        this.sendChallenge(session, 'ticket', {})
      } else {
        this.sendAbort(session, 'wamp.error.authorization_failed')
      }
    } else {
      this.getRouter().getRealm(session.realmName, function (realm) {        
        realm.joinSession(session)
        this.sendWelcome(ctx)
        if (message.will) {
          session.setDisconnectPublish(ctx, this.makePublishCmd(ctx, message.will))
        }
      }.bind(this))
    }
  }

  sendWelcome (ctx) {
    // Return Code
    // 0 	Connection accepted
    // 1 	Connection refused, unacceptable protocol version
    // 2 	Connection refused, identifier rejected
    // 3 	Connection refused, server unavailable
    // 4 	Connection refused, bad user name or password
    // 5 	Connection refused, not authorized
    ctx.mqttSend({ returnCode: 0, sessionPresent: false, cmd: 'connack' })
  }

  makePublishCmd(ctx, message) {
    let opt = {}
    if (message.retain) {
      opt.retain = true
    }
    if (message.qos >= 1) {
      opt.history = true
    }

    let uri = mqttParse(message.topic)
    this.checkAuthorize(ctx, 'publish', uri, message.messageId)

    let data
    if (message.payload.length === 0) {
      data = null
    } else {
      data = { payload: message.payload }
    }
    return {
      mtype: 'publish',
      uri,
      data,
      id: message.messageId,
      opt
    }  
  }

  handle (ctx, session, msg) {
    if (typeof msg !== 'object') {
      this._router.emit(SESSION_WARNING, session, 'no object arrived', msg)
      ctx.mqttClose(1003, 'protocol violation')
      return
    }
    let mtype = msg.cmd
    if (!handlers[mtype]) {
      this._router.emit(SESSION_WARNING, session, 'command not found', msg)
      ctx.mqttClose(1003, 'protocol violation')
      return
    }
    try {
      handlers[mtype].call(this, ctx, session, msg)
    } catch (err) {
      if (err instanceof RealmError) {
        ctx.mqttClose(1003, err.message)
        console.log(err)
      } else {
        throw err
      }
    }
  }

  getProtocol () {
    return 'mqtt'
  }
}

handlers.connect = function (ctx, session, message) {
  if (session.realm === null) {
    this.hello(ctx, session, message)
  } else {
    ctx.mqttClose(1002, 'protocol violation')
  }
  return false
}

handlers.disconnect = function (ctx, session, message) {
  // do not send WILL notification at correct disconnect
  session.cleanDisconnectPublish()
  return false
}

handlers.publish = function (ctx, session, message) {
  this.checkRealm(session)
  // Error: Invalid publish topic 'com.myapp/#', does it contain '+' or '#'?

  let cmd = this.makePublishCmd(ctx, message)
  if (message.qos >= 1) {
    cmd.ack = true
  }
  session.realm.doPush(ctx, cmd)
}

cmdAck.publish = function (cmd) {
  this.mqttSend({ cmd: 'puback', messageId: cmd.id })
}

handlers.pingreq = function (ctx, session, message) {
  ctx.mqttSend({ cmd: 'pingresp' })
}

handlers.subscribe = function (ctx, session, message) {
  this.checkRealm(session)
  let pkg = {
    id: message.messageId,
    granted: [],
    count: message.subscriptions.length
  }
  ctx.setId(pkg)
  for (let index=0; index < message.subscriptions.length; index++) {
    let qos = Math.min(message.subscriptions[index].qos, 1)
    pkg.granted[index] = qos
    let uri = mqttParse(message.subscriptions[index].topic)
    if (this.checkAuthorize(ctx, 'subscribe', uri, index)) {
      let opt = {}
      if (qos > 0) {
        opt.keepHistoryFlag = true
      }
      session.realm.doTrace(ctx, {
        mtype: 'subscribe',
        id: index,
        uri,
        opt
      })
    }
  }
}

cmdAck.subscribe = function (cmd) {
  this.mqttSubscribeDone()
}

module.exports = MqttGate
